/**************************************************************************
***
***    Copyright (c) 1995-1997 by ABKGroup, UCLA VLSI CAD Laboratory,
***    UCLA Computer Science Department, Los Angeles, CA 90095-1596 USA,
***    and by the Regents of the University of California.
***    All rights reserved.
***
***    No portion of this file may be used, copied, or transmitted for
***    any purpose, including any research or commercial purpose,
***    without the express written consent of Professor Andrew B. Kahng,
***    UCLA Computer Science Department, 3713 Boelter Hall, Los Angeles,
***    CA 90095-1596 USA.   Any copy made of this file must include this
***    notice.    ABKGroup and the Regents of the University of California
***    retain title to all copies of this file and/or its compiled
***    derivatives.
***
***************************************************************************/

#include <fstream.h>
#include "ABKCommon/abkcommon.h"
#include "DB/DB.h"
#include "RBPlaceFromDB.h"

int main(int argc, const char *argv[])
{
  StringParam auxFileName("f",argc,argv);
  abkfatal(auxFileName.found(),"Usage: prog -f filename.aux");

  BoolParam check("checkDB",argc,argv);
  DB::Parameters dbParams;
  dbParams.alwaysCheckConsistency = check;

  DB db(auxFileName,dbParams);

  db.saveDEF("placed.def",true);
  ofstream eaOut("easyAccess.out");
  eaOut<<db.easyAccess;
  eaOut.close();

  RBPlaceParams rbParams(argc,argv);
  RBPlaceFromDB rbpl(db, rbParams);
  ofstream rbOut("rbplacemt.out");
  rbOut << rbpl;
  rbOut.close();

  Placement pl(1);
  db >> pl;

  cout<<" Total CCs       "<< db.getNetlist().getNumCCs()       << endl;
  cout<<" Total SCs       "<< db.getNetlist().getNumSCs()       << endl;
  cout<<" Total CoreCells "<< db.getNetlist().getNumCoreCells() << endl;
  cout<<" Total AllPads   "<< db.getNetlist().getNumAllPads()   << endl;
  cout<<" Total IOPads    "<< db.getNetlist().getNumIOPads()    << endl;
  cout<<" Total IOPins    "<< db.getNetlist().getNumIOPins()    << endl;


  return 0;
}
