/**************************************************************************
***
***    Copyright (c) 1995-1997 by ABKGroup, UCLA VLSI CAD Laboratory,
***    UCLA Computer Science Department, Los Angeles, CA 90095-1596 USA,
***    and by the Regents of the University of California.
***    All rights reserved.
***
***    No portion of this file may be used, copied, or transmitted for
***    any purpose, including any research or commercial purpose,
***    without the express written consent of Professor Andrew B. Kahng,
***    UCLA Computer Science Department, 3713 Boelter Hall, Los Angeles,
***    CA 90095-1596 USA.   Any copy made of this file must include this
***    notice.    ABKGroup and the Regents of the University of California
***    retain title to all copies of this file and/or its compiled
***    derivatives.
***
***************************************************************************/

#include "ABKCommon/abkcommon.h"
#include "DB/DB.h"
#include "RBPlaceFromDB.h"

using std::vector;

int main(int argc, const char *argv[])
{
  StringParam auxFileName("f",argc,argv);
  abkfatal(auxFileName.found(),"Usage: prog -f filename.aux");

  BoolParam check("checkDB",argc,argv);
  DB::Parameters dbParams;
  dbParams.alwaysCheckConsistency = check;

  DB db(auxFileName,dbParams);

  cout<<" Building RBPlacement"<<endl; 
  RBPlacement::Parameters rbParams(argc,argv);
  RBPlaceFromDB rbp(db, rbParams);

  cout << " Row-based placement ";
  if (rbp.isConsistent()) cout << "consistent " << endl;
  else cout << "inconsistent " << endl;

  itRBPCoreRow crItr;
  //for(crItr = rbp.coreRowsBegin(); crItr != rbp.coreRowsEnd(); crItr++)
  crItr = rbp.coreRowsBegin();
  {
    itRBPSubRow srItr;
    //for(srItr = crItr->subRowsBegin(); srItr != crItr->subRowsEnd(); srItr++)
    srItr = crItr->subRowsBegin();
    {
	vector<unsigned> cellIds;
	vector<unsigned> siteSpacing;
	vector<double>   spacing;

	cout<<"Original SR: "<<*srItr<<endl<<endl;

	RBPSubRow& sr = const_cast<RBPSubRow&>(*srItr);


	//________pick up all the cells, and space them uniformly_____
	cout<<"________pick up all the cells, and space them uniformly"<<endl;

	cout<<" Extracting all cells"<<endl;
	rbp.extractCellsFromSR(sr,cellIds,siteSpacing);
	cout<<" SR After Extraction: "<<sr<<endl;
	cout<<"  Num Cells:   "<<cellIds.size()<<endl;
	cout<<" SiteSpacing:  "<<siteSpacing<<endl;

	cout<<" Embedding cells in original order - equal spacing"<<endl;
	rbp.embedCellsInSR(sr,cellIds);
	cout<<" SR After Embedding: "<<sr<<endl;

	//____________look at the order...see that its uniform__
	cout<<"__________look at the order...see that its uniform"<<endl;

	cout<<" Extracting all cells - siteSpacing requested"<<endl;
	rbp.extractCellsFromSR(sr,cellIds,siteSpacing);
	cout<<" SR After Extraction: "<<sr<<endl;
	cout<<"  Num Cells:   "<<cellIds.size()<<endl;
	cout<<" Spacing:      "<<siteSpacing<<endl;

	cout<<" Embedding cells in original order"<<endl;
	rbp.embedCellsInSR(sr,cellIds,siteSpacing);
	cout<<" SR After Embedding: "<<sr<<endl;

	//____________extract the cells, and get the loc offsets__
	cout<<"________extract the cells, and get the loc offsets"<<endl;

	cout<<" Extracting all cells - spacing requested"<<endl;
	rbp.extractCellsFromSR(sr,cellIds,spacing);
	cout<<" SR After Extraction: "<<sr<<endl;
	cout<<"  Num Cells:   "<<cellIds.size()<<endl;
	cout<<" Spacing:      "<<spacing<<endl;

	cout<<" Embedding cells in original order - with spacings"<<endl;
	rbp.embedCellsInSR(sr,cellIds,spacing);
	cout<<" SR After Embedding: "<<sr<<endl;

	//____________check that the cells are still uniformly spaced__
	cout<<"________check that the cells are still uniformly spaced"<<endl;

	cout<<" Extracting all cells - siteSacing requested"<<endl;
	rbp.extractCellsFromSR(sr,cellIds,siteSpacing);
	cout<<" SR After Extraction: "<<sr<<endl;
	cout<<"  Num Cells:   "<<cellIds.size()<<endl;
	cout<<" Spacing:      "<<siteSpacing<<endl;
    }
  }

  return 0;
}

