/**************************************************************************
***    
***    Copyright (c) 1995-1997 by ABKGroup, UCLA VLSI CAD Laboratory,
***    UCLA Computer Science Department, Los Angeles, CA 90095-1596 USA, 
***    and by the Regents of the University of California.    
***    All rights reserved.   
***
***    No portion of this package may be used, copied, or transmitted for
***    any purpose, including any research or commercial purpose,
***    without the express written consent of Professor Andrew B. Kahng,
***    UCLA Computer Science Department, 3713 Boelter Hall, Los Angeles,
***    CA 90095-1596 USA.   Any copy made of this file must include this
***    notice.    ABKGroup and the Regents of the University of California
***    retain title to all copies of this file and/or its compiled
***    derivatives.
***
***************************************************************************/

//990112	aec	pulled from violationReport.cxx

#ifndef __VIOLRPT_SUPPORT_H__
#define __VIOLRPT_SUPPORT_H__

#include <functional>
#include "ABKCommon/abkcommon.h"
#include "Placement/placement.h"
#include "DB/DB.h"
#include "RBPlace/RBPlacement.h"
#include "violationReport.h"


/******** function signatures **************/

// The function below called by applyToRegion (therefore not a member of
// violationReport.
       void logError1 (unsigned&, unsigned, std::vector <Violations>&,
                      const Violations&, unsigned, const dbNet&, const char *);
 
static bool isLegal(const Symmetry& cellSym, const Orientation& cellOri);
 
static bool isLegal (const Orientation& rowOri,
                     const Symmetry& siteSym,
                     const Orientation& cellOri);
 
/********* function definitions *************/


/*
   This class lets us binary search on vectors of cell indices, to
   find those in an x-coordinate region, where the x-bound may
   not be the location of a known cell.
*/
class cellXcompare
{
public:  // these are public so mapping function can change them...
  const Placement* _pl;
  double _ref;
public:
  cellXcompare (const Placement* pl, double ref) { _pl = pl; _ref = ref; }
  bool operator() (unsigned index1, unsigned index2)
  {
    return 
      ((index1 != InvalidUnsigned) ? (*_pl)[index1].x : _ref)
        < ((index2 != InvalidUnsigned) ? (*_pl)[index2].x : _ref);
  }
};

template<class kind>
class unaryPredicate : public std::unary_function<kind, bool>
{
public:
  bool operator() (kind) {return true;}
};


class applyToRegionPred : public unaryPredicate<unsigned>
{
public:
  virtual bool operator() (unsigned) = 0;
};

class wireTestFunctor : public applyToRegionPred
{
protected:
  const dbLayerGeom& _lGeom;
  const DB& _db;
  unsigned _verbose;
  std::vector <Violations>& _violations;
  unsigned& _errors;
  const dbNet& _snet;
  const ViolationReport & _violReport;
  
public:
  
  wireTestFunctor (const DB& db, 
                   unsigned& errors,
                   unsigned verbose,
                   std::vector <Violations>& violations,
                   const dbLayerGeom& lGeom,
                   const dbNet& snet,
                   const ViolationReport& violReport)
    : _lGeom(lGeom), _db(db), _verbose(verbose), _violations(violations), 
      _errors(errors), _snet(snet), _violReport(violReport)
  {}
  
  bool operator() (unsigned cellID)
  {
    if (_violReport.checkOverlap(cellID, _lGeom))
      logError1(_errors, _verbose, _violations, Violations("wOverlap"),
               cellID, _snet, "Cell overlaps a special net");
    return true;
  }
};


#endif
